//# This file is a part of marzer/soagen and is subject to the the terms of the MIT license.
//# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
//# See https://github.com/marzer/soagen/blob/master/LICENSE for the full license text.
//# SPDX-License-Identifier: MIT
#pragma once

#define SOAGEN_VERSION_MAJOR 0
#define SOAGEN_VERSION_MINOR 3
#define SOAGEN_VERSION_PATCH 0
#define SOAGEN_VERSION_STRING "0.3.0"
